<?php

declare(strict_types=1);

namespace Drupal\lms\Controller;

use Drupal\Core\Url;
use Drupal\Core\Utility\Error;
use Drupal\group\Entity\GroupInterface;
use Drupal\lms\Entity\Bundle\Course;
use Drupal\lms\Entity\CourseStatus;
use Drupal\lms\Entity\LessonStatus;
use Drupal\lms\Exception\TrainingException;

/**
 * Common methods for training navigation logic classes.
 */
trait CourseControllerTrait {

  /**
   * Helper function that gets field item deltas for lesson and activity.
   */
  private function getDeltas(CourseStatus $course_status): array {
    $lesson_status = $course_status->get('current_lesson_status')->entity;
    $lesson_delta = NULL;
    $group = $course_status->get('gid')->entity;
    foreach ($group->get(Course::LESSONS) as $delta => $item) {
      if ($lesson_status->get('lesson')->target_id === $item->target_id) {
        $lesson_delta = $delta;
        break;
      }
    }
    if ($lesson_delta === NULL) {
      throw new TrainingException(NULL, NULL, TrainingException::LESSON_REMOVED);
    }

    return [
      'lesson_delta' => $lesson_delta,
      'activity_delta' => (int) $lesson_status->get('current_activity')->value,
    ];
  }

  /**
   * Error handler.
   */
  private function handleError(GroupInterface $group, \Exception $e): Url {
    // Training exception - reset current lesson and activity and redirect
    // to the lesson.
    if ($e instanceof TrainingException) {
      $iterations = 0;
      // Loop could be used in case of free navigation, otherwise there'll
      // always be a single iteration without additional exceptions.
      do {
        $success = TRUE;
        $course_status = $e->getLpStatus();
        $step = $e->getStep();
        if ($course_status === NULL || $step === NULL) {
          if ($e->getCode() === TrainingException::NO_LESSONS) {
            $this->messenger()->addError($this->t('This learning path has no lessons.'));
          }
          else {
            $this->messenger()->addError($this->t('An unexpected error occurred.'));
          }
          return Url::fromRoute('entity.group.canonical', [
            'group' => $group->id(),
          ]);
        }
        try {
          if ($e->getCode() === TrainingException::REQUIRED_NOT_FINISHED) {
            $lesson_status = $this->trainingManager->loadLessonStatus($course_status->id(), $step->target_id);
            $message = $this->t('You need to finish lesson "@lesson" first.', [
              '@lesson' => $step->entity->label(),
            ]);
          }
          elseif ($e->getCode() === TrainingException::REQUIRED_NOT_STARTED) {
            $lesson_status = $this->trainingManager->initializeLesson($course_status, $step);
            $message = $this->t('You need to finish lesson "@lesson" first.', [
              '@lesson' => $step->entity->label(),
            ]);
          }
          // In case of insufficient score the entire lesson needs to be
          // restarted.
          elseif ($e->getCode() === TrainingException::INSUFFICIENT_SCORE) {
            $lesson_status = $this->trainingManager->initializeLesson($course_status, $step);
            $message = $this->t('Insufficient score on "@lesson" to proceed. Please retake the lesson.', [
              '@lesson' => $step->entity->label(),
            ]);
          }

          $lesson_status->save();
          $course_status->set('current_lesson_status', $lesson_status);
          $course_status->save();
          $this->messenger()->addError($message);
          return Url::fromRoute('lms.group.answer_form', [
            'group' => $group->id(),
          ] + $this->getDeltas($course_status));
        }
        catch (TrainingException $nested_exception) {
          $success = FALSE;
          $e = $nested_exception;
        }

        $iterations++;
        if ($iterations > 50) {
          throw new \Exception('Too many iterations, please inspect the code.');
        }
        // First lesson always can be initialized so no risk of infinitely
        // throwing a TrainingException.
      } while (!$success);
    }

    // Other exception type - redirect to the group, log exception and
    // display a generic message.
    $this->messenger()->addError($this->t('An error occurred while trying to start training. Please report the issue to your site administrator.'));
    Error::logException($this->getLogger('lms'), $e, 'Training @group_id start error for user @user_id. Message: @message Backtrace: @backtrace_string', [
      '@group_id' => $group->id(),
      '@user_id' => $this->currentUser()->id(),
      '@message' => $e->getMessage(),
    ]);
    return Url::fromRoute('entity.group.canonical', [
      'group' => $group->id(),
    ]);
  }

  /**
   * Lesson status summarry getter.
   */
  public function getLessonStatusSummary(LessonStatus $lesson_status) {
    return $this->t('@lesson (completed @completed / @total activities, @result)', [
      '@lesson' => $lesson_status->getLesson()->label(),
      '@completed' => $lesson_status->getAnswerCount(),
      '@total' => $lesson_status->getQuestionCount(),
      '@result' => $lesson_status->isEvaluated() ? $this->t('score: @score% (@required% required)', [
        '@score' => $lesson_status->getScore(),
        '@required' => $lesson_status->getRequiredScore(),
      ]) : $this->t('not all evaluated'),
    ]);
  }

}
