<?php

declare(strict_types=1);

namespace Drupal\lms;

use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\lms\Entity\Activity;
use Drupal\lms\Entity\Answer;
use Drupal\lms\Entity\Bundle\Course;
use Drupal\lms\Entity\CourseStatus;
use Drupal\lms\Entity\Lesson;
use Drupal\lms\Entity\LessonStatus;
use Drupal\lms\Exception\TrainingException;
use Drupal\lms\Plugin\Field\FieldType\LMSReferenceItem;
use Symfony\Component\DependencyInjection\Attribute\Autowire;
use Symfony\Component\DependencyInjection\Attribute\AutowireCallable;

/**
 * The Training manager.
 */
class TrainingManager {

  /**
   * The constructor.
   */
  public function __construct(
    #[AutowireCallable(service: 'entity_type.manager', method: 'getStorage')]
    protected \Closure $getStorage,
    #[AutowireCallable(service: 'datetime.time', method: 'getRequestTime')]
    private \Closure $getRequestTime,
    #[Autowire(service: 'logger.channel.training_manager', lazy: TRUE)]
    protected readonly LoggerChannelInterface $logger
    ) {}

  /**
   * Gets a learning path status entity given a learning path and user.
   *
   * @param \Drupal\lms\Entity\Bundle\Course $training
   *   The training.
   * @param \Drupal\Core\Session\AccountInterface $account
   *   The user account.
   * @param mixed[] $conditions
   *   Array of query conditions.
   *
   * @return \Drupal\lms\Entity\CourseStatus|null
   *   LP status entity or NULL.
   */
  public function loadCourseStatus(Course $training, AccountInterface $account, array $conditions = []): ?CourseStatus {
    $course_status_storage = ($this->getStorage)('lms_course_status');
    $query = $course_status_storage->getQuery()
      ->accessCheck(FALSE)
      ->condition('gid', $training->id())
      ->condition('uid', $account->id());

    foreach ($conditions as $property => $value) {
      $query->condition($property, $value);
    }
    $query->sort('started', 'DESC');
    $query->range(0, 1);
    $result = $query->execute();

    if ($result) {
      return $course_status_storage->load(reset($result));
    }
    return NULL;
  }

  /**
   * Get a count of certificates for a user.
   *
   * @todo Will this be used anywhere?
   */
  public function getCertificateCount(AccountInterface $account): int {
    $course_status_storage = ($this->getStorage)('lms_course_status');
    $results = $course_status_storage->getQuery()
      ->accessCheck(FALSE)
      ->condition('uid', $account->id())
      ->condition('status', 'passed')
      ->exists('gid.entity.field_certificate')
      ->condition('gid.entity.status', 1)
      ->groupBy('gid')
      ->execute();
    $statuses = $course_status_storage->loadMultiple($results);

    $count = 0;
    foreach ($statuses as $status) {
      // Check if the user is still a member of private groups.
      $group = $status->gid->entity;
      if ($group->isPrivate() && $group->getMember($account) === NULL) {
        continue;
      }
      $count++;
    }

    return $count;
  }

  /**
   * Get a count of passed trainings for a user.
   *
   * @todo Will this be used anywhere?
   */
  public function getPassedTrainingsCount(AccountInterface $account): int {
    return (int) ($this->getStorage)('lms_course_status')->getAggregateQuery()
      ->accessCheck(FALSE)
      ->condition('uid', $account->id())
      ->condition('status', 'passed')
      ->condition('gid.entity.status', 1)
      ->groupBy('gid')
      ->count()
      ->execute();
  }

  /**
   * Helper method to get the first LP step.
   */
  public function getFirstStep(Course $group): LMSReferenceItem {
    return $group->get(Course::LESSONS)->first();
  }

  /**
   * Code saver.
   */
  public function loadLessonStatus(string $course_status_id, string $lesson_id): ?LessonStatus {
    $statuses = $this->loadLessonStatusMultiple($course_status_id, [$lesson_id]);
    if (\count($statuses) === 0) {
      return NULL;
    }
    return \reset($statuses);
  }

  /**
   * Code saver.
   *
   * @return \Drupal\lms\Entity\LessonStatus[]
   *   User lesson status array.
   */
  public function loadLessonStatusMultiple(string $course_status_id, array $lesson_ids): array {
    $lesson_status_storage = ($this->getStorage)('lms_lesson_status');
    $query = $lesson_status_storage->getQuery()
      ->accessCheck(FALSE)
      ->condition('course_status', $course_status_id)
      ->condition('lesson', $lesson_ids, 'IN');
    $results = $query->execute();
    return $lesson_status_storage->loadMultiple($results);
  }

  /**
   * Helper method to get all training steps ordered.
   *
   * No branching support.
   *
   * @return array
   *   Steps data.
   */
  public function getOrderedSteps(Course $group): array {
    $steps = $group->get(Course::LESSONS);
    $result = [];
    foreach ($steps as $step_delta => $step) {
      $lesson = $step->entity;
      $randomization = $lesson->getRandomization();
      $result[$step->target_id] = [
        'label' => $lesson->label(),
        'delta' => $step_delta,
        'required_score' => $step->getSuccessScoreMin(),
        'activities' => [],
      ];

      // If we don't have randomization, activities are not attempt-specific
      // and can be populated here as they are identical as in lesson statuses.
      if ($randomization === 0) {
        $activity_items = $lesson->get(Lesson::ACTIVITIES);
        foreach ($activity_items as $activity_delta => $activity_item) {
          $result[$step->target_id]['activities'][$activity_item->target_id] = [
            'label' => $activity_item->entity->label(),
            'delta' => $activity_delta,
            'activity' => $activity_item->entity,
          ];
        }
      }
    }

    return $result;
  }

  /**
   * Helper method to get next lesson for CourseStatus.
   */
  public function getNextLessonStatus(CourseStatus $course_status): ?LessonStatus {
    $current_lesson_status_field = $course_status->get('current_lesson_status');
    if ($current_lesson_status_field->isEmpty()) {
      return NULL;
    }

    $current_lesson_id = $current_lesson_status_field->entity->get('lesson')->target_id;
    $course = $course_status->getCourse();
    foreach ($course->get(Course::LESSONS) as $delta => $step) {
      if ($current_lesson_id === $step->target_id) {
        $next_delta = $delta + 1;
        $next_step = $course->get(Course::LESSONS)->get($next_delta);
        if ($next_step === NULL) {
          return NULL;
        }
        $lesson_status = $this->loadLessonStatus($course_status->id(), $next_step->target_id);
        if ($lesson_status !== NULL) {
          return $lesson_status;
        }
        return $this->initializeLesson($course_status, $next_step);
      }
    }

    return NULL;
  }

  /**
   * Get previous lesson reference item.
   */
  private function getPreviousStep(CourseStatus $course_status, string $lesson_id): ?LMSReferenceItem {
    $steps = $course_status->get('gid')->entity->get(Course::LESSONS);
    $previous_step = NULL;
    foreach ($steps as $step) {
      if ($step->target_id === $lesson_id) {
        return $previous_step;
      }
      $previous_step = $step;
    }

    return NULL;
  }

  /**
   * Gets backwards navigation parameters for CourseStatus.
   *
   * @return string[]
   *   Keyed array with 2 string values:
   *     - lesson_delta (NULL if going back 1 activity),
   *     - activity_delta (NULL if going back to the previous lesson).
   */
  public function getBackNavParameters(Answer $answer): array {
    $course_status = $answer->getLessonStatus()->get('course_status')->entity;
    $lesson_status = $course_status->get('current_lesson_status')->entity;
    if ($lesson_status === NULL) {
      return [];
    }

    $activity_delta = $lesson_status->getCurrentActivityDelta();
    if ($activity_delta === NULL) {
      return [];
    }

    $current_lesson_id = $lesson_status->get('lesson')->target_id;
    $lesson_delta = NULL;
    $steps = $course_status->get('gid')->entity->get(Course::LESSONS);

    if ($activity_delta === 0) {
      // Access check.
      if (!$course_status->allowAnyActivity()) {
        return [];
      }

      $previous_lesson_delta = NULL;
      foreach ($steps as $delta => $step) {
        if ($step->target_id === $current_lesson_id) {
          break;
        }
        $previous_lesson_delta = $delta;
      }

      if ($previous_lesson_delta === NULL) {
        return [];
      }
      $lesson_delta = $previous_lesson_delta;

      // Set activity delta to NULL so the next method knows we're
      // going back one lesson.
      $activity_delta = NULL;
    }
    else {
      // Access check.
      $lesson = $lesson_status->get('lesson')->entity;
      if (
        !$lesson->getBackwardsNavigation() &&
        !$course_status->allowAnyActivity()
      ) {
        return [];
      }

      foreach ($steps as $delta => $step) {
        if ($step->target_id === $current_lesson_id) {
          $lesson_delta = $delta;
          break;
        }
      }

      $activity_delta = $activity_delta - 1;
    }

    return [
      'lesson_delta' => $lesson_delta,
      'activity_delta' => $activity_delta,
    ];
  }

  /**
   * Helper method to go backwards in a learning path.
   *
   * @return mixed[]
   *   Array of redirect route options.
   */
  public function goBackwards(CourseStatus $course_status, $back_parameters): ?array {
    if (\count($back_parameters) === 0) {
      return NULL;
    }

    if ($back_parameters['activity_delta'] !== NULL) {
      $lesson_status = $course_status->get('current_lesson_status')->entity;
      $lesson_status->set('current_activity', $back_parameters['activity_delta']);
      $lesson_status->save();
    }
    elseif ($back_parameters['lesson_delta'] !== NULL) {
      $group = $course_status->get('gid')->entity;
      $step = $group->get(Course::LESSONS)->get($back_parameters['lesson_delta']);
      $lesson_status = $this->getLessonStatus($course_status, $step->entity);
      $back_parameters['activity_delta'] = $lesson_status->getLastActivityDelta();
      $lesson_status->set('current_activity', $back_parameters['activity_delta']);
      $lesson_status->save();
      $course_status->set('current_lesson_status', $lesson_status);
      $course_status->save();
    }

    return [
      'group' => $course_status->get('gid')->target_id,
    ] + $back_parameters;
  }

  /**
   * Perform all operations required to start a training.
   *
   * @return mixed[]
   *   Array of redirect route options.
   */
  public function initializeTraining(Course $group, AccountInterface $user): CourseStatus {
    $current_step = $group->get(Course::LESSONS)->first();
    if (!$current_step instanceof LMSReferenceItem) {
      throw new TrainingException(NULL, NULL, TrainingException::NO_LESSONS);
    }

    // Create a LP status entity.
    $course_status = ($this->getStorage)('lms_course_status')->create([
      'gid' => $group->id(),
      'uid' => $user->id(),
      'last_activity_ts' => ($this->getRequestTime)(),
      'finished' => 0,
      'status' => CourseStatus::STATUS_PROGRESS,
    ]);
    $course_status->save();

    // It should always be possible to initialize the first lesson (no
    // possible parent lesson requirement).
    $lesson_status = $this->initializeLesson($course_status, $current_step);
    $lesson_status->save();
    $course_status->set('current_lesson_status', $lesson_status);
    $course_status->save();

    return $course_status;
  }

  /**
   * Checks if a lesson can be initialized.
   *
   * @throws \Drupal\lms\Exception\TrainingException
   */
  private function canTakeLesson(CourseStatus $course_status, string $lesson_id): void {
    $previous_step = $this->getPreviousStep($course_status, $lesson_id);
    // No previous step - we can always initialize the first lesson.
    if ($previous_step === NULL) {
      return;
    }

    $mandatory = $previous_step->isMandatory();
    $min_score = $previous_step->getSuccessScoreMin();
    if (!$mandatory && $min_score === 0) {
      return;
    }

    $previous_lesson = $previous_step->entity;

    $lesson_status = $this->loadLessonStatus($course_status->id(), $previous_lesson->id());
    if ($lesson_status === NULL) {
      throw new TrainingException($course_status, $previous_step, TrainingException::REQUIRED_NOT_STARTED);
    }
    if ($mandatory && !$lesson_status->isFinished()) {
      throw new TrainingException($course_status, $previous_step, TrainingException::REQUIRED_NOT_FINISHED);
    }

    // @todo Technically we should check if the user actually can
    // complete the lesson assumming all remaining questions are evaluated.
    // For this not to cause performance issues we'd need to calculate
    // current score and add max possible score for each unevaluated
    // activity. That'd need to be implemented in self::updateLessonStatus()
    // and saved to some new lesson status field.
    if (!$lesson_status->isEvaluated()) {
      return;
    }

    if ($lesson_status->getScore() < $min_score) {
      throw new TrainingException($course_status, $previous_step, TrainingException::INSUFFICIENT_SCORE);
    }
  }

  /**
   * Initialize lesson.
   *
   * Creates a new lesson attempt linked to a CourseStatus.
   */
  public function initializeLesson(CourseStatus $course_status, LMSReferenceItem $step, ?Activity $activity = NULL): LessonStatus {

    // Check if requirements are met (previous lesson required score,
    // field_proceed_on_evaluated).
    $this->canTakeLesson($course_status, $step->target_id);

    $lesson_status = $this->loadLessonStatus($course_status->id(), $step->target_id);
    // If there already was an attempt, restart.
    if ($lesson_status !== NULL) {
      // @todo Save lesson attempt data for analytical purposes or push to
      // a LRS when needed before deletion.
      $answers = ($this->getStorage)('lms_answer')->loadByProperties([
        'lesson_status' => $lesson_status->id(),
      ]);
      foreach ($answers as $answer) {
        $answer->delete();
      }
      $lesson_status->delete();
    }

    // Create new lesson status.
    $lesson_status = ($this->getStorage)('lms_lesson_status')->create([
      'course_status' => $course_status,
      'lesson' => $step->target_id,
      'required_score' => $step->getSuccessScoreMin(),
      'finished' => 0,
    ]);

    $this->setLessonStatusActivities($lesson_status);

    return $lesson_status;
  }

  /**
   * Sets activities and max score of a lesson status.
   */
  public function setLessonStatusActivities(LessonStatus $lesson_status, ?string $current_activity_delta = NULL) {
    $lesson = $lesson_status->get('lesson')->entity;
    if ($lesson === NULL) {
      throw new \Exception(sprintf('User lesson status references invalid lesson %d.', $lesson_status->get('lesson')->target_id));
    }
    $activity_items = $lesson->get(Lesson::ACTIVITIES);

    if (\count($activity_items) === 0) {
      throw new \Exception(sprintf('Lesson "%s" (%d) with no activities attempted.', $lesson->label(), $lesson->id()));
    }

    $randomization = $lesson->getRandomization();
    if ($randomization > 0) {
      // Convert to an array so shuffle can be used.
      $as_array = [];
      foreach ($activity_items as $delta => $item) {
        $as_array[$delta] = $item;
      }
      $activity_items = $as_array;
      unset($as_array);

      \shuffle($activity_items);
      if ($randomization === 2) {
        $random_activities = $lesson->get('random_activities')->value;
        $activity_items = \array_slice($activity_items, 0, (int) $random_activities);
      }
    }

    $activity_ids = [];
    $max_score = 0;
    foreach ($activity_items as $item) {
      $activity_ids[] = $item->target_id;
      $max_score += $item->getMaxScore();
    }

    $lesson_status->setActivityIds($activity_ids);
    $lesson_status->set('total_questions', \count($activity_ids));
    $lesson_status->setMaxScore($max_score);

    if ($current_activity_delta === NULL) {
      $current_activity_delta = 0;
    }
    $lesson_status->set('current_activity', $current_activity_delta);
  }

  /**
   * Get learning path status.
   *
   * @return array<CourseStatus, LessonStatus, string>
   *   Array of activity form route parameters.
   */
  public function getCurrentCourseStatus(Course $group, AccountInterface $user): CourseStatus {
    $course_status = $this->loadCourseStatus($group, $user, $group->getInitialLpStatusConditions());

    if ($course_status === NULL) {
      return $this->initializeTraining($group, $user);
    }
    return $course_status;
  }

  /**
   * Load answer entity for a specified lesson status and activity.
   */
  public function loadAnswer(LessonStatus $lesson_status, Activity $activity): ?Answer {
    $answer_storage = ($this->getStorage)('lms_answer');
    $query = $answer_storage->getQuery();
    $result = $query
      ->accessCheck(FALSE)
      ->condition('lesson_status', $lesson_status->id())
      ->condition('activity', $activity->id())
      ->execute();
    return \count($result) === 0 ? NULL : $answer_storage->load(\reset($result));
  }

  /**
   * Get answer entity for a specified lesson status and activity.
   */
  public function createAnswer(LessonStatus $lesson_status, Activity $activity): Answer {
    return ($this->getStorage)('lms_answer')->create([
      // @todo Remove some useless fields.
      // @see \Drupal\lms\Entity\Answer::baseFieldDefinitions().
      'type' => $activity->bundle(),
      'user_id' => $lesson_status->get('course_status')->entity->getUserId(),
      'activity' => $activity->id(),
      'lesson' => $lesson_status->get('lesson')->target_id,
      'lesson_status' => $lesson_status->id(),
      'score' => 0,
    ]);
  }

  /**
   * Get max score for an activity.
   */
  public function getActivityMaxScore(Lesson $lesson, Activity $activity): int {
    foreach ($lesson->get(Lesson::ACTIVITIES) as $item) {
      if ($item->target_id === $activity->id()) {
        return $item->getMaxScore();
      }
    }
    return 0;
  }

  /**
   * Get answer plugin ID for a given activity.
   */
  public function getActivityPluginId(Activity $activity): ?string {
    $activity_type_id = $activity->bundle();
    $activity_type = ($this->getStorage)('lms_activity_type')->load($activity_type_id);
    if ($activity_type === NULL) {
      return NULL;
    }
    return $activity_type->get('pluginId');
  }

  /**
   * Helper method to get answers by activity ID from a lesson status.
   *
   * @param array<string> $activity_ids
   *   Validation purpose - array of activity IDs that should belong to
   *   the current lesson.
   */
  public function getAnswersByActivityId(LessonStatus $lesson_status, ?array $activity_ids = NULL): array {
    $answers = ($this->getStorage)('lms_answer')->loadByProperties([
      'lesson_status' => $lesson_status->id(),
    ]);
    $answers_by_activity_id = [];
    foreach ($answers as $answer) {
      $activity_id = $answer->get('activity')->target_id;
      // Dangling references in field items case.
      if ($activity_ids !== NULL && !\array_key_exists($activity_id, $activity_ids)) {
        $this->logger->warning('Answer (ID: %answer_id) to a activity (ID: %activity_id) that does not exist in lesson (ID: %lesson_id) found.', [
          '%answer_id' => $answer->id(),
          '%activity_id' => $activity_id,
          '%lesson_id' => $lesson_status->get('lesson')->target_id,
        ]);
        continue;
      }
      $answers_by_activity_id[$activity_id] = $answer;
    }
    return $answers_by_activity_id;
  }

  /**
   * Check if all answers for a lesson status are evaluated.
   */
  public function allAnswersEvaluated(LessonStatus $lesson_status): bool {
    $query = ($this->getStorage)('lms_answer')->getQuery();
    $count = $query
      ->accessCheck(FALSE)
      ->condition('lesson_status', $lesson_status->id())
      ->condition('evaluated', FALSE)
      ->count()
      ->execute();
    return $count === 0;
  }

  /**
   * Check if all lesson statuses for a course status are evaluated.
   */
  public function allLessonStatusesEvaluated(CourseStatus $course_status): bool {
    $query = ($this->getStorage)('lms_lesson_status')->getQuery();
    $count = $query
      ->accessCheck(FALSE)
      ->condition('course_status', $course_status->id())
      ->condition('evaluated', FALSE)
      ->count()
      ->execute();
    return $count === 0;
  }

  /**
   * Update user lesson status data.
   *
   * @return bool
   *   Did the user lesson status change?
   */
  public function updateLessonStatus(LessonStatus $lesson_status): bool {
    $activity_ids = $lesson_status->getActivityIds();
    $activities = ($this->getStorage)('lms_activity')->loadMultiple($activity_ids);
    $activities_count = \count($activities);
    if ($activities_count === 0) {
      throw new \Exception(sprintf('Trying to calculate lesson status %d for a lesson with no activities.', $lesson_status->id()));
    }

    $answers_by_activity_id = $this->getAnswersByActivityId($lesson_status, $activity_ids);

    $max_score = $lesson_status->getMaxScore();
    $all_answered = TRUE;
    $all_evaluated = TRUE;
    $activities_count = 0;
    $answered_count = 0;
    $evaluated_count = 0;
    $obtained_score = 0;
    foreach ($activities as $activity) {
      $activities_count++;
      if (!\array_key_exists($activity->id(), $answers_by_activity_id)) {
        $all_answered = FALSE;
        $all_evaluated = FALSE;
        continue;
      }

      $answer = $answers_by_activity_id[$activity->id()];
      $answered_count++;
      if (!$answer->isEvaluated()) {
        $all_evaluated = FALSE;
        continue;
      }

      $evaluated_count++;
      $obtained_score += $answer->getScore();
    }

    if ($max_score === 0) {
      $percentage = $evaluated_count / $activities_count;
    }
    else {
      $percentage = $obtained_score / $max_score;
    }
    $percentage = (int) \round($percentage * 100);

    if (
      (int) $lesson_status->get('given_answers')->value === $answered_count &&
      $lesson_status->getScore() === $percentage &&
      $lesson_status->isEvaluated() === $all_evaluated
    ) {
      return FALSE;
    }

    // Set lesson status values.
    $lesson_status->set('given_answers', $answered_count);
    $lesson_status->setScore($percentage);

    if ($all_answered && !$lesson_status->isFinished()) {
      $lesson_status->setFinished(($this->getRequestTime)());
      $lesson_status->get('current_activity')->delete();
    }

    // Calculate passed percentage.
    $pass_percentage = \round($answered_count / $activities_count * 100);
    $lesson_status->set('percent', $pass_percentage);

    $lesson_status->setEvaluated($all_evaluated);
    $lesson_status->save();
    return TRUE;
  }

  /**
   * Set last activity timestamp on LP status entity.
   */
  public function setLastActivityTime(CourseStatus $course_status): void {
    $course_status->setLastActivity(($this->getRequestTime)());
  }

  /**
   * Calculate Course status data.
   *
   * @param \Drupal\lms\Entity\CourseStatus $course_status
   *   The Course Status entity to update.
   * @param bool $last_activity
   *   Did the user just completed the last activity?
   */
  public function updateCourseStatus(CourseStatus $course_status, bool $last_activity = FALSE) {
    $course = $course_status->getCourse();
    $lesson_ids = [];

    $steps = $course->get(Course::LESSONS);
    foreach ($steps as $step) {
      $lesson_ids[$step->target_id] = $step->target_id;
    }
    $lessons = ($this->getStorage)('lms_lesson')->loadMultiple($lesson_ids);

    $lesson_statuses = ($this->getStorage)('lms_lesson_status')->loadByProperties([
      'course_status' => $course_status->id(),
    ]);
    $lesson_statuses_by_lesson_id = [];
    foreach ($lesson_statuses as $status) {
      $lesson_id = $status->get('lesson')->target_id;
      // Data corruption case or recalculating finished attempts of modified
      // learning paths which also should never happen.
      // @todo Throw an exception?
      if (!\array_key_exists($lesson_id, $lessons)) {
        continue;
      }
      $lesson_statuses_by_lesson_id[$lesson_id] = $status;
    }

    $all_finished = TRUE;
    $all_mandatory_answered = TRUE;
    $all_mandatory_evaluated = TRUE;
    $all_mandatory_completed = TRUE;
    $max_score = 0;
    $score_weighted_sum = 0;
    foreach ($steps as $step) {
      $lesson_id = $step->target_id;
      $lesson = $step->entity;
      // @todo Use count query in a separate method.
      $lesson_questions_count = $lesson->get(Lesson::ACTIVITIES)->count();
      if ($lesson->getRandomization() === 2) {
        $random_activities = (int) $lesson->get('random_activities')->value;
        if ($random_activities < $lesson_questions_count) {
          $lesson_questions_count = $random_activities;
        }
      }

      // Lesson not started.
      if (!\array_key_exists($lesson_id, $lesson_statuses_by_lesson_id)) {
        $all_finished = FALSE;
        if ($step->isMandatory()) {
          $all_mandatory_answered = FALSE;
          $all_mandatory_evaluated = FALSE;
          $all_mandatory_completed = FALSE;
        }
        continue;
      }

      $lesson_status = $lesson_statuses_by_lesson_id[$lesson_id];
      // Lesson not finished.
      if (!$lesson_status->isFinished()) {
        $all_finished = FALSE;
        if ($step->isMandatory()) {
          $all_mandatory_answered = FALSE;
          $all_mandatory_evaluated = FALSE;
          $all_mandatory_completed = FALSE;
        }
        continue;
      }

      $lesson_score = $lesson_status->getScore();
      $lesson_attempt_max_score = $lesson_status->getMaxScore();
      $max_score += $lesson_attempt_max_score;
      $score_weighted_sum += $lesson_score * $lesson_attempt_max_score;
      if ($step->isMandatory()) {
        if (!$lesson_status->isEvaluated()) {
          $all_mandatory_evaluated = FALSE;
          $all_mandatory_completed = FALSE;
        }
        elseif ($lesson_score < $lesson_status->getRequiredScore()) {
          $all_mandatory_completed = FALSE;
        }
      }
    }

    if (!$course_status->isFinished()) {
      if (
        $all_finished ||
        // Set LP as finished if the user just answered the last activity
        // and all mandatory actifities have been finished (branching or
        // free navigation case).
        ($last_activity && $all_mandatory_answered)
      ) {
        $course_status->setFinished(($this->getRequestTime)());
      }
    }

    // Determine the status of the learning path.
    // LP can be finished regardless of this method logic in a branching
    // scenario.
    if ($course_status->isFinished()) {
      if ($all_mandatory_completed) {
        $status = CourseStatus::STATUS_PASSED;
      }
      elseif (!$all_mandatory_evaluated) {
        $status = CourseStatus::STATUS_NEEDS_EVALUATION;
      }
      else {
        $status = CourseStatus::STATUS_FAILED;
      }
    }
    else {
      $status = CourseStatus::STATUS_PROGRESS;
    }

    // Revisiting a training - don't update status if previously passed.
    $no_fail_mode_passed = (
      $course_status->getStatus() === CourseStatus::STATUS_PASSED &&
      $course_status->get('gid')->entity->revisitMode()
    );
    if (!$no_fail_mode_passed) {
      $course_status->setStatus($status);
    }

    // Set score if course is finished and evaluated.
    if ($course_status->isFinished() && $all_mandatory_evaluated) {
      if ($max_score === 0) {
        $course_status->setScore(0);
      }
      else {
        $course_status->setScore((int) \round($score_weighted_sum / $max_score));
      }
    }
    else {
      $course_status->setScore(NULL);
    }

    $course_status->save();
    // @todo Add percent complete field to CourseStatus and calculate basing on
    // total questions and answered questions count here to optimize some
    // views.
  }

  /**
   * Defines the training status cache tag.
   */
  public static function trainingStatusTag(string $group_id, string $user_id): string {
    return 'training_status:' . $group_id . ':' . $user_id;
  }

  /**
   * Calculate user course data.
   *
   * Utility method for updating corrupt data (should be used in an post
   * update hook, hence the ID parameters).
   *
   * @return bool
   *   Was the recalculation successful?
   */
  public function updateUserCourseData(string $group_id, string $user_id): bool {

    $group = ($this->getStorage)('group')->load($group_id);
    if ($group === NULL || !$group instanceof Course) {
      return FALSE;
    }
    $user = ($this->getStorage)('user')->load($user_id);
    if ($user === NULL) {
      return FALSE;
    }

    // Recalculate all status entities for this user and group.
    $course_statuses = ($this->getStorage)('lms_course_status')->loadByProperties([
      'uid' => $user->id(),
      'gid' => $group->id(),
    ]);
    if (\count($course_statuses) === 0) {
      return FALSE;
    }

    foreach ($course_statuses as $course_status) {
      $lesson_statuses = ($this->getStorage)('lms_lesson_status')->loadByProperties([
        'course_status' => $course_status->id(),
      ]);

      // Update lesson statuses.
      $status_data = [];
      foreach ($lesson_statuses as $lesson_status) {
        $this->updateLessonStatus($lesson_status);
        $status_data[$lesson_status->id()] = [
          'percent' => $lesson_status->get('percent')->value,
          'started' => $lesson_status->getCreatedTime(),
        ];
      }

      // There should be only one per CourseStatus so delete all but the best or
      // the latest one.
      if (\count($status_data) > 1) {
        $preserve_id = '';
        $max_percentage = 0;
        foreach ($status_data as $id => $item) {
          if ($item['percent'] > $max_percentage) {
            $preserve_id = $id;
          }
        }
        if ($preserve_id === '') {
          $latest_ts = 0;
          foreach ($status_data as $id => $item) {
            if ($item['started'] > $latest_ts) {
              $preserve_id = $id;
            }
          }
        }

        // If $preserve_id is still set to an empty string, looks like
        // we can delete all anyway.
        foreach ($lesson_statuses as $id => $lesson_status) {
          if ($id === $preserve_id) {
            continue;
          }
          $lesson_status->delete();
        }
      }

      // Recalculate LP status.
      $this->updateCourseStatus($course_status);
    }
  }

  /**
   * Reset training.
   *
   * Development and data fixing purpose.
   *
   * @todo Use for the case of modifications to courses that are already
   * in progress by some students.
   */
  public function resetTraining(string $group_id, ?string $user_id = NULL): void {
    $course_status_storage = ($this->getStorage)('lms_course_status');

    $course_status_query = $course_status_storage->getQuery();
    $course_status_query
      ->accessCheck(FALSE)
      ->condition('gid', $group_id);
    if ($user_id !== NULL) {
      $course_status_query->condition('uid', $user_id);
    }

    $course_status_ids = $course_status_query->execute();

    // If user ID is not provided and some course statuses are
    // completed - only clear current lesson and current activity values
    // if students wanted to revisit the training.
    if ($user_id === NULL) {
      foreach ($course_status_storage->loadMultiple($course_status_ids) as $course_status) {
        if (
          $course_status->getStatus() === CourseStatus::STATUS_PROGRESS ||
          $course_status->getStatus() === CourseStatus::STATUS_NEEDS_EVALUATION
        ) {
          continue;
        }
        $course_status->set('current_lesson_status', NULL);
        $course_status->save();
        unset($course_status_ids[$course_status->id()]);
      }
    }

    // Lesson statuses and answers will be handled in entity delete hooks.
    if (\count($course_status_ids) !== 0) {
      foreach ($course_status_storage->loadMultiple($course_status_ids) as $course_status) {
        $course_status->delete();
      }
    }
  }

}
